import os
import bibtexparser
import bibtexparser.customization
from . import tags2table as t2t
from . import table as t

def bib2labels(
    bib_file,
    keyword_filter = "",
    keyword_separator = ":",
    label_rename = {
        "ID": "reference"
    },
    fields = ["ID"]):
    """
    Extracts a labels dict suitable for table generation from a bibtex reference database
    bib_file           -- path to bibtex file
    keyword_filter     -- will skip keywords that don't begin with the keyword_filter text
    keyword_separator  -- character used to delimit hierarchical keyword
    label_rename       -- dictionary mapping old name to new name
    fields             -- additional bibtex fields to extract in addition to keywords
    output_file        -- filename of output table
    returns            -- labels dict
    """
    with open(bib_file) as bibtex_file:
        bibtex_str = bibtex_file.read()

    def customizations(record):
        # bibtexparser customizations
        # convert latex special characters (e.g. {\"a})
        record = bibtexparser.customization.convert_to_unicode(record)
        # turn keywords field into a list of keywords
        record = bibtexparser.customization.keyword(record)
        return record
    
    parser = bibtexparser.bparser.BibTexParser()
    parser.customization = customizations
    bib_database = bibtexparser.loads(bibtex_str, parser=parser)
    entries = bib_database.entries
    
    rows = []
    cols_set = set()
    
    for entry in entries:
        row = {}
        
        # Extract keywords
        keywords = entry['keyword']
        for keyword in keywords:
            if not keyword.startswith(keyword_filter):
                continue
            sub_keywords = keyword.split(keyword_separator)
            sub_keywords = [label_rename.get(k, k) for k in sub_keywords]
            if len(sub_keywords) > 1:
                head = sub_keywords[0]
                tail = sub_keywords[1:]
                row[head] = tail
                cols_set.add(head)
            else:
                row[keyword] = True
                cols_set.add(keyword)
        
        # extract extra fields
        for field in fields:
            field_rename = label_rename.get(field, field)
            row[field_rename] = entry[field]
            cols_set.add(field_rename)
        
        rows.append(row)
    
    cols = sorted(cols_set)
    
    # leave rest to inference
    lables_dict = {
        # TODO: Leave col structure to auto-inference.
        #       atomatically group mutually exculsive cols
        #       into hierarchies.
        'cols': cols,
        # TODO: Allow way to specify that source col should
        #       be sorted last (perhaps a "sort-hint" col attribute)
        # TODO: Allow types to be dict so that we can
        #       specify ref type without needing to know
        #       how many columns in advance
        #'types': {
        #    label_rename.get("ID", "ID"): "ref"
        #},
        'sort_rows': True,
        'data': rows,
    }
    
    return lables_dict

def labels2txt(
    labels,
    output_file):
    """
    Generate plaintext table
    labels      -- labels dict
    output_file -- filename of output table
    """
    table = t2t.tags2table(labels)
    presenter = t.TxtTable()
    txt = presenter.present(table)
    with open(output_file, 'w') as out:
        out.write(txt)

#def labels2tsv(
#    labels,
#    output_file):
#    """
#    Generate tab separated value table
#    labels      -- labels dict
#    output_file -- filename of output table
#    """
#    pass

#def labels2latex(
#    labels,
#    output_file,
#    output_wrapper,
#    bib_file):
#    """
#    Generate LaTeX table
#    labels         -- labels dict
#    output_file    -- filename of output LaTeX table
#    output_wrapper -- filename of output LaTeX wrapper to compile table
#    bib_file       -- bib filename for LaTeX wapper to use for citations keys
#    """
#    pass



from . import uniquebool

def create_matrix(n_rows, n_cols, fill):
    # creates a 2d array of fill
    # n_rows -- number of rows
    # n_cols -- number of cols
    # fill -- value to initialize each cell with
    # returns -- list of lists
    result = []
    for i in range(0, n_rows):
        row = [fill] * n_cols
        result.append(row)
    return result

class Table:
    def __init__(self):
        self.n_cols = 0
        self.n_data_rows = 0
        self.n_header_rows = 0
        self.head = []
        # keeps traked of merged cells in header
        self.head_stretch = []
        self.data = []
        self.data_indent = []
        self.built = False

    # Table construction
    def set_cols(self, n_cols):
        self.n_cols = n_cols

    def set_data_rows(self, n_rows):
        self.n_data_rows = n_rows

    def set_header_rows(self, n_rows):
        self.n_header_rows = n_rows

    def build(self):
        # There are two types of empty-like cells:
        # * None (displayed as '-')
        # * Invisible (empty string)
        self.head = create_matrix(self.n_header_rows, self.n_cols, '')
        self.head_stretch = create_matrix(self.n_header_rows, self.n_cols, 1)
        self.data = create_matrix(self.n_data_rows, self.n_cols, '')
        self.data_indent = create_matrix(self.n_data_rows, self.n_cols, 0)
        self.built = True

    def set_row_cell(self, r, c, txt, indent=0):
        assert self.built
        self.data[r][c] = txt
        self.data_indent[r][c] = indent

    def set_header_cell(self, r, start_col, end_col, txt):
        """
        start_col -- start of merged cols (included)
        end_col -- end of merged cols (excluded)
        """
        assert self.built
        self.head[r][start_col] = txt
        self.head_stretch[r][start_col] = end_col - start_col

    # Table viewing
    def get_dims(self):
        assert self.built
        return ((self.n_header_rows, self.n_data_rows), self.n_cols)

    def iter_rows(self):
        assert self.built
        for r in self.data:
            yield r

class TableFormatter:
    """
    Graphically/Textually presents the data in a table
    """
    pass

class TxtTable(TableFormatter):
    class DimensionedTable:
        def __init__(self, table):
            # copy table functionality
            self.iter_rows = table.iter_rows
            self.get_dims = table.get_dims

            # copy table properties
            self.n_cols = table.n_cols
            self.n_data_rows = table.n_data_rows
            self.n_header_rows = table.n_header_rows
            self.head = table.head
            self.head_stretch = table.head_stretch
            self.data = table.data
            self.data_indent = table.data_indent
            self.built = table.built

            # own properties
            self.col_width = []
            for col in range(self.n_cols):
                # initialize col widths to 0
                self.col_width.append(0)

            self.min_col_right_pos = [] # minimum nuber of chars right side of col must end at.
                                        # measured as absolute position from leftmost part of table.
            for col in range(self.n_cols):
                # initialize col widths to 0
                self.min_col_right_pos.append(0)
        def get_header(self, hr, c):
            return self.display(self.head[hr][c])
        def get_cell(self, r, c):
            return self.display(self.data[r][c])
        def display(self, cell_data):
            return TxtTable._display(cell_data)

    def __init__(self):
        pass

    def present(self, table):
        """
        table -- table.Table
        returns -- table formated as text
        """
        dims = TxtTable.DimensionedTable(table)

        left_pos = 0
        # find minimum size of each column (greedily take smallest possible, starting from leftmost column)
        for c in range(0, dims.n_cols):
            for hr in range(0, dims.n_header_rows):
                cell_width = len(dims.get_header(hr, c))
                c_right_edge = c + dims.head_stretch[hr][c] - 1
                min_cell_right_pos = left_pos + cell_width
                curr = dims.min_col_right_pos[c_right_edge]
                dims.min_col_right_pos[c_right_edge] = max(curr, min_cell_right_pos)

            for r in range(dims.n_data_rows):
                cell_width = len(dims.get_cell(r, c))
                min_cell_right_pos = left_pos + cell_width
                curr = dims.min_col_right_pos[c]
                dims.min_col_right_pos[c] = max(curr, min_cell_right_pos)

            dims.col_width[c] = dims.min_col_right_pos[c] - left_pos
            left_pos = dims.min_col_right_pos[c] + 1 # include 1 char padding

        result = ""
        num_breaks = max(0, dims.n_cols - 1)
        col_total_width = sum(dims.col_width)
        table_width = col_total_width + num_breaks
        result += "=" * table_width + "\n"
        for hr in range(dims.n_header_rows):
            padded_row = []
            underlines = ""
            underline_upto = 0
            c = 0
            while c < dims.n_cols:
                stretch = dims.head_stretch[hr][c]
                contents = dims.get_header(hr, c)
                
                # any non-leaf header should be underlined, even if it only stretches over one cell
                if hr < dims.n_header_rows - 1 and contents != '':
                    # underline all cols in stretched/merged cols
                    stretch_total_width = sum(dims.col_width[c:c+stretch])
                    num_breaks = stretch - 1
                    col_width = stretch_total_width + num_breaks
                    underlines += "-" * col_width

                    padded_row.append(TxtTable._pad(contents, col_width))

                    c += stretch
                    if c < dims.n_cols:
                        underlines += " " # col separation
                else:
                    col_width = dims.col_width[c]
                    underlines += " " * col_width # no underline

                    padded_row.append(TxtTable._pad(contents, col_width))

                    c += 1
                    if c < dims.n_cols:
                        underlines += " " # col separation

            result += " ".join(padded_row) + "\n"
            if hr < dims.n_header_rows - 1: # don't attempt to underline final row
                result += underlines + "\n"
        result += "=" * table_width + "\n"
        for r in range(dims.n_data_rows):
            padded_row = []
            for c in range(dims.n_cols):
                contents = dims.get_cell(r, c)
                col_width = dims.col_width[c]
                padded_row.append(TxtTable._pad(contents, col_width))
            result += " ".join(padded_row) + "\n"
        result += "=" * table_width
        return result

    @classmethod
    def _pad(cls, s, pad_length):
        remainder = pad_length - len(s)
        assert remainder >= 0
        return s + " " * remainder

    @classmethod
    def _display(cls, td, indent=0):
        """
        td -- data to display
        returns -- string
        """
        
        # Need the types, else key False == key 0 and key True == key 1.
        conv_map = {
            (uniquebool.UniqueBool, uniquebool.TRUE): 'Y',
            (uniquebool.UniqueBool, uniquebool.FALSE): 'N',
            # Type of None is NoneType, but NoneType isn't exposed in Python3
            (type(None), None): '-'
        }
        
        if type(td) is list:
            if len(td) == 0:
                # empty list, will be represented as '-'
                td = None
            else:
                # hierarchy, display final element with indent
                indent += len(td) - 1
                td = td[-1]
        
        result = " " * indent
        key = (type(td), td)
        if key in conv_map:
            td = conv_map[key]
            result += td
        else:
            result += str(td)
        
        return result

    def cmp(self, txta, txtb):
        """
        Compares textually formatted tables for equivlence.
        Tables with different data or col names are not the same.
        Tables with extra whitespace/streched column widths
        may be accptable.

        returns -- true if tables match.
        """
        # todo: allow more flexibility
        
        # ignore any extra whitespace at end of each line
        txta = '\n'.join([l.rstrip() for l in txta.split('\n')])
        txtb = '\n'.join([l.rstrip() for l in txtb.split('\n')])
        
        return txta.strip() == txtb.strip()



import copy
from . import table as t
from . import uniquebool

def indent(s, amount=2):
    """
    Indents a block of text (useful for debug purposes).
    Not intended for use in indenting table cells!
    """
    results = []
    for line in s.split('\n'):
        results.append(" "*amount + line)
    return "\n".join(results)

class Node(object):
    def __init__(self, name):
        self.name = name
        self.children = []
        self.descendants = 1 # count leaf nodes as 1
        self.depth = 1
        self.parent = None
    
    def add_child(self, child):
        self.children.append(child)
        child.parent = self
    
    @property
    def child_count(self):
        return len(self.children)
    
    def update_descendants(self):
        """Updates descendant count (assumes children already updated)"""
        descendants = sum([c.descendants for c in self.children])
        self.descendants = max(1, descendants) # if leaf node, count as 1.
        self.depth = 1 + max([c.depth for c in self.children]) # max depth
    
    def recalc_descendants(self):
        """Recursively updates descendant counts (assumes acylic graph)"""
        # back up tree (in reverse order - from leaves up)
        branches = []
        sweep = [self]
        while sweep:
            sweep_next = []
            for node in sweep:
                if len(node.children) > 0:
                    branches.append(node)
                    sweep_next += node.children
            sweep = sweep_next
        
        for node in reversed(branches):
            node.update_descendants()
    
    @property
    def chain(self):
        chain = []
        node = self
        # chain back to, but not including, root node
        while node and node.parent:
            chain.append(node.name)
            node = node.parent
        chain = reversed(chain)
        return tuple(chain)
    
    def pad(self, length=None):
        """Deprecated"""
        if length == None:
            # default: pad all children to same depth
            length = self.depth
            
        # Pads all children chains with empty nodes to ensure all chains are
        # same target length.
        if length == 1:
            # When target length reaches 0, then we shouldn't have any children
            assert self.children == []
            return
            
        if self.children == []:
            # If no children, extend
            self.add_child(Node(None))

        for c in self.children:
            # Recursively pad children
            c.pad(length - 1)
    
    def __str__(self):
        result = [str(self.name)]
        result += [indent(str(c)) for c in self.children]
        return '\n'.join(result)

class HeaderNode(Node):
    pass
    
class DataNode(Node):
    def __init__(self, *args, **kwargs):
        super(DataNode, self).__init__(*args, **kwargs)
        self.row_start = None
        self.height = 1 # default partition height
    
    @property
    def val(self):
        # alias for name of node
        return self.name

def create_header_tree(cols):
    root = HeaderNode('Root')
    
    # broad pass top-down to construct tree
    current_sweep = [(root, col) for col in cols]
    while current_sweep:
        next_sweep = []

        for r, c in current_sweep:
            if type(c) is tuple:
                # sub-col c contains futher branches
                name, tail = c
                node = HeaderNode(name)
                r.add_child(node)
                
                if type(tail) is list:
                    # c hss multiple branches
                    for child in tail:
                        next_sweep.append((node, child))
                else:
                    # c has only one branch
                    next_sweep.append((node, tail))
            elif type(c) is list:
                raise Exception('Expected tuple or string')
            else:
                # c is leaf
                node = HeaderNode(c)
                r.add_child(node)

        current_sweep = next_sweep
    
    root.recalc_descendants() # todo, make Node class automatically update counts.
    
    # root with all elements connected, and descendants counted
    return root

def set_headers(cols, data, table):
    header_tree = create_header_tree(cols)
    table_cols = header_tree.descendants
    num_header_rows = header_tree.depth - 1 # don't include root node
    
    # Depth first search for child nodes.
    # note that it is important that cols come out in correct order!
    leaves = []
    
    def visit(r, c, node):
        # r,c r,c args will be oriented wrong way,
        # but we only care about the node
        if not node.children:
            leaves.append(node)
    
    # walk tree and fill table cells
    walk_tree(
        header_tree, # walk_tree is intended for walking row_tree,
                     # but can also walk col_headers
                     # (but visit r,c args will be oriented wrong way)
        visit
    )
    
    # set cols to just the final leaves
    col_chains = [leaf.chain for leaf in leaves]
    
    return header_tree, col_chains, table_cols, num_header_rows
    
def fill_headers(tree, table):
    """
    tree -- header_tree returned from set_headers
    table -- built Table
    return -- col_leaves
    """

    leaves = []
    
    # sweep tree top to bottom, left to right.
    depth = 0
    current_sweep = tree.children
    while current_sweep:
        cols_to_left = 0
        next_sweep = []
        for node in current_sweep:
            end_col = cols_to_left + node.descendants           
            if depth < tree.depth - 1 - node.depth:
                # column hierarchy is shorter than others
                # add vertical padding and come back next sweep
                next_sweep.append(node)
                cols_to_left = end_col
            else:
                txt = node.name
                table.set_header_cell(depth, cols_to_left, end_col, txt)
                cols_to_left = end_col
                for child in node.children:
                    next_sweep.append(child)
                if not node.children:
                    leaves.append(node)
        depth += 1
        current_sweep = next_sweep
    
    # leaves are col names
    return leaves

def create_data_tree(cols, data):
    """
    cols -- col_chains returned by set_headers
    data -- tags data
    """
    root = DataNode('Root')
    root.row_start = 0
    root.height = len(data) # root partition contains all data rows
                            # this will later be sub-partitioned
    sweep = [root]
    
    for c, col_name in enumerate(cols):
        # Split up data into horizontal partitions.
        # Consecutive rows with the same col value will be placed in the same partition
        next_sweep = []
        
        for partition in sweep:
            rstart = partition.row_start
            rend   = rstart + partition.height
            
            sub_partition = None
            hierarchy_context = [] # Current hierarchy context.
                                   # Used to decide whether to insert
                                   # special hierarchy partition row
            
            for r in range(rstart, rend):
                rdata = data[r]

                # lookup cell data for this row, column
                try:
                    cdata = fuzzy_row_lookup(rdata, col_name)
                except KeyError:
                    cdata = None
                
                if type(cdata) is list:
                    # hierarchy
                    diff_index = 0
                    
                    while diff_index < min(len(cdata)-1, len(hierarchy_context)):
                        if cdata[diff_index] == hierarchy_context[diff_index]:
                            diff_index += 1
                        else:
                            # diff_index will be the first index where
                            # desired hierarchy context differs from last
                            # hierarchy context
                            break
                    
                    for level in cdata[diff_index:-1]:
                        # insert header for level
                        # Start a new partition for this row
                        sub_partition = DataNode(level)
                        sub_partition.row_start = r
                        sub_partition.height = 0
                        partition.add_child(sub_partition)
                        next_sweep.append(sub_partition)
                    
                    hierarchy_context = cdata[:-1]
                
                # every row in last column is always its own partition
                # (prevents creating rows that are completely blank / missing).
                if sub_partition != None and cdata == sub_partition.val and c != len(cols)-1:
                    # Repeated value in this column.
                    # Merge this cell into the last group.
                    sub_partition.height += 1
                else:
                    # Start a new partition for this row
                    sub_partition = DataNode(cdata)
                    sub_partition.row_start = r
                    partition.add_child(sub_partition)
                    next_sweep.append(sub_partition)
        
        sweep = next_sweep
    
    root.recalc_descendants() # todo, make Node class automatically update counts.

    # root node with all partitions attached
    return root

def setup_data_cels(col_chains, data):
    """
    col_chains -- col_chains leaves returned by set_headers
    data -- tags data
    retrn -- data_tree, num_rows
    """
    data_tree = create_data_tree(col_chains, data)
    num_rows = data_tree.descendants
    return data_tree, num_rows

def fill_data_cels(table, row_tree):
    """
    table    -- built table
    row_tree -- return value of setup_data_cels
    """
    def visit(r, c, node):
        table.set_row_cell(r, c, node.val)
        
    # walk tree and fill table cells
    walk_tree(
        row_tree,
        visit
    )

def walk_tree(tree, visit_func):
    """
    Depth-first walk of tree of Nodes.
    r,c indexes are under assumption of walking row_tree from left to right
    (to remove duplicate rows).
    tree -- root Node
    visit_func -- lambda r, c, node
    """
    # walk the tree
    r = 0
    c = 0
    # path depth -> list of children
    unvisited = [] # list of list of children seen at each node in path.
                  # TODO: Use doubly linked list for efficiency.
    unvisited.append(list(tree.children))

    while True:
        #print 'walk {}'.format((r,c))
        
        unvisited_children = unvisited[-1]
        if len(unvisited_children) > 0:
            # dive in
            node = unvisited_children.pop(0)
            unseen = list(node.children)
            unvisited.append(unseen)
            
            # visit code here
            visit_func(r, c, node)
            # end visit code
            
            # track position
            c += 1
        else:
            # backtrack
            while len(unvisited_children) == 0:
                unvisited.pop()
                
                if len(unvisited) == 0:
                    # seen everything
                    return
                
                unvisited_children = unvisited[-1]
                # track position
                c -= 1
            
            # track position
            r += 1

def fuzzy_row_key(rdata, col):
    """
    Lookup correct index using closest col name.
    If the col has trailing Nones, these are ignored
    return -- key
    """
    # Todo: Normalize rdata first
    
    try:
        # Remove trailing Nones.
        col = col[:col.index(None)]
    except ValueError:
        # No trailing Nones. Leave col as is.
        pass
    
    if len(col) == 1:
        # For non-hierarchical cols, allow just string (e.g. "Sweet")
        # rather than packing as tupple.
        col = col[0]
    
    return col
    
def fuzzy_row_lookup(rdata, col):
    """
    Lookup row data using closest col name.
    If the col has trailing Nones, these are ignored
    """
    # Todo: Normalize rdata first
    
    col = fuzzy_row_key(rdata, col)

    # will throw a KeyError if not present
    cdata = rdata[col]
    return cdata

def normalize_table(col_chains, data, types):
    """
    Normalize data in table to make processing simpler
    col_chains -- from set_headers
    data -- table input data array
    types -- table input type array
    return -- data_norm, the normalized data array
    """
    # Currently:
    # * Normalize hierarchies (ensure consistent level depths)
    # Todo: fill in missing cells
    
    data_norm = copy.deepcopy(data)
    
    # Set missing cells in bool cols to False
    for c, col_name in enumerate(col_chains):
        if types[c] == 'bool':
            for r, rdata in enumerate(data):
                
                col_norm = fuzzy_row_key(rdata, col_name)
                
                try:
                    cdata = rdata[col_norm]
                except KeyError:
                    # missing data => convert to False
                    cdata = uniquebool.FALSE
                
                data_norm[r][col_norm] = cdata
    
    # Normalize hierarchies
    for c, col_name in enumerate(col_chains):
        hierarchy_headers = set()
        
        # find hierarchical cols
        for r, rdata in enumerate(data):
            
            try:
                cdata = fuzzy_row_lookup(rdata, col_name)
            except KeyError:
                cdata = None
            
            if type(cdata) is list:
                key = tuple(cdata)
                hierarchy_headers.add(key)
        
        # old header -> new length
        header_lengths = dict()
        
        # find ideal header lengths
        for header in hierarchy_headers:
            # pad any main headers above this header
            # to be at least the length of this header.
            # e.g., if there is a A.B header,
            # then push any data for the A header into A.-
            sub_headers = []
            
            for i in range(1, len(header)+1):
                sub_header = header[0:i]
                sub_headers.append(sub_header)
            
            for sub_header in sub_headers:
                if sub_header not in header_lengths:
                    header_lengths[sub_header] = 0 # initialize

                header_lengths[sub_header] = max(header_lengths[sub_header], len(header))
        
        # pad headers
        for r, rdata in enumerate(data):
            
            col_norm = fuzzy_row_key(rdata, col_name)
            
            try:
                cdata = rdata[col_norm]
            except KeyError:
                cdata = None

            if type(cdata) is list:
                key = tuple(cdata)
                
                if key in header_lengths:
                    pad_len = header_lengths[key]
                    # pad header with Nones to make correct length
                    pad_diff = pad_len - len(cdata)
                    assert pad_diff >= 0
                    cdata_new = cdata + [None] * pad_diff                    
                    #print ('old {} new {}'.format(cdata, cdata_new))
                    
                    # use row_new
                    data_norm[r][col_norm] = cdata_new
    
    return data_norm

def infer_type(col_chains, types, data):    
    for c, col_name in enumerate(col_chains):
        if types[c] != None:
            # already set
            continue
        
        bool_count = 0
        empty_count = 0
        other_count = 0

        # find hierarchical cols
        for r, rdata in enumerate(data):
            try:
                cdata = fuzzy_row_lookup(rdata, col_name)
            except KeyError:
                cdata = None
            
            if cdata == None:
                empty_count += 1
            elif type(cdata) is uniquebool.UniqueBool:
                bool_count += 1
            else:
                other_count += 1
        
        if other_count > 0:
            # non-boolean. Assume text.
            # Empty => '-'
            types[c] = 'str'
        else:            
            # Default to assuming bool
            # Empty => 'F'
            types[c] = 'bool'

    return types

def tags2table(table_arg):
    """
    See examples for how to specify table_arg
    table_arg -- dict of data and cols
    return -- Table
    """
    table = t.Table()
    data = table_arg['data']
    cols = table_arg['cols']
    
    # Replace pesky True, False objects with our own uniquebool.TRUE, uniquebool.FALSE
    # objects. This prevents partitioning issues due to True == 1.
    data = uniquebool.deep_replace_bool(data)

    header_tree, col_chains, table_cols, num_header_rows = set_headers(cols, data, table)

    if not 'types' in table_arg:
        table_arg['types'] = [None] * table_cols
    
    types = table_arg['types']
    assert len(types) == table_cols
    # Attempt to infer unspecified types from data
    types = infer_type(col_chains, types, data)
    
    data = normalize_table(col_chains, data, types)
    row_tree, num_rows = setup_data_cels(col_chains, data)

    table.set_cols(len(col_chains))
    table.set_header_rows(num_header_rows)
    table.set_data_rows(num_rows)
    table.build()

    fill_headers(header_tree, table)
    fill_data_cels(table, row_tree)

    return table



#!/usr/bin/env python

from setuptools import setup

with open('README.rst') as f:
    readme = f.read()

setup(
    name='labels2tables',
    version='0.1.0',
    description="Turns bibtex keywords into an academic summary table",
    long_description=readme,
    author='Andrew Simmons',
    author_email='anjsimmo@gmail.com',
    url='https://github.com/anjsimmo/labels2tables',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Topic :: Text Processing :: Markup',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    packages=['labels2tables'],
    install_requires=[
        'bibtexparser',
        'enum34'
    ],
    test_suite="tests"
)



import copy
import collections
from enum import Enum

# Python's built in bools, True and False, are equal to 1 and 0.
# This can lead to many subtle bugs in table columns that mix bools with numbers.
# For example, if one row starts with a True cell, and the row below starts with a 1 cell,
# these might be accidentally merged together in a partition because True == 1.
# To solve this, we introduce our own bool variables that behave like ordinary objects.

class UniqueBool(object):
    def __init__(self, name):
        self._name = name
    def __repr__(self):
        return self._name
    def __deepcopy__(self, _):
        # For equity to work as expected, we want to keep the same instance
        # http://stackoverflow.com/questions/9887501/deepcopy-does-not-respect-metaclass
        return self
    def __copy__(self, _):
        # For equity to work as expected, we want to keep the same instance
        # http://stackoverflow.com/questions/9887501/deepcopy-does-not-respect-metaclass
        return self

TRUE = UniqueBool("TRUE") # unique object
FALSE = UniqueBool("FALSE") # unique object

def deep_replace_bool(obj):
    """
    obj -- list, tuple or item that contains ordinary Python bools.
           Assumed to be acyclic.
    return -- santitized copy of object with Python bools replaced by TRUE and FALSE objects.
              Only values will be replaced, keys will be unchanged.
              Immutable collections (other than tuple) are not supported
    """
    def rep(item):
        if item is True:
            return TRUE
        elif item is False:
            return FALSE
        else:
            return item
    
    cpy = copy.deepcopy(obj)
    deep_replace(cpy, rep)
    return cpy

def deep_replace(obj, replace_func):
    """
    obj -- list, tuple or item. Assumed to be acyclic.
    replace_func -- function(item) to return replacement items
    """
    if isinstance(obj, collections.Mapping):
        for k,v in obj.items():
            obj[k] = deep_replace(v, replace_func)
        return obj
    elif isinstance(obj, collections.MutableSequence):
        for k,v in enumerate(obj):
            obj[k] = deep_replace(v, replace_func)
        return obj
    elif type(obj) is tuple:
        return tuple(deep_replace(v, replace_func) for v in obj)
    else:
        return replace_func(obj)
